%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% 1. Vizsgld meg az albbi Prolog programot, amely az N-vezr feladatot
%    oldja meg "generate and test" mdszerrel! rj egy hatkonyabb vltozatot,
%    queens_c/2 nven, a korutinszervezsi lehetsgek felhasznlsval!

% ================= Original code =============

% queens(+N, ?Q): Q is a *safe* placement of queens, i.e. no two queens
% attack each other.
queens(N, Qs) :-
        length(Qs, N),
	place_queens(Qs, N),
        safe(Qs).

% List Qs, of length N, describes a placement of queens on a chessboard of
% size N*N: if Q[i] = j, then a queen is placed in row i, column j.
place_queens(Qs, N) :-
	(   foreach(Q, Qs), param(N)
	do  between(1, N, Q)
	).

% Expanding the do-loop the following recursive version of the above
% predicate is obtained.
/*
place_queens([], _).
place_queens([Q|Qs], N) :-
 	between(1, N, Q),
	place_queens(Qs, N).
*/

% safe(Qs): List Qs describes a safe placement of queens.
safe([]).
safe([Q|Qs]):- 
        no_attack(Qs, Q, 1),
	safe(Qs).

% no_attack(Qs, Q, I): A quuen placed in row i, position Q does not attack
% any of the queens Qs[1], Qs[2], etc, placed in rows i+I, i+I+1, etc.
no_attack([], _, _).
no_attack([X|Xs], Y, I):-
      no_threat(X, Y, I),
      J is I+1,
      no_attack(Xs, Y, J).

% Queens placed in columns X and Y in rows of distance I do not attack each
% other. A bit more formally: queens placed in row i, column X and in row
% i+I, column Y do not attack each other.
no_threat(X, Y, I) :-
        Y =\= X, Y =\= X-I, Y =\= X+I.

% between(+M, +N, ?X): X is an integer such that M =< X =< N, where M and N
% are integers.
between(M, N, M):-
	M =< N.
between(M, N, I):-
	M < N,
	M1 is M+1,
	between(M1, N, I).

% ================= Code with coroutining =============

% queens_c(+N, ?Q): Q is a *safe* placement of queens, i.e. no two queens
% attack each other.
queens_c(N, Qs) :-
        length(Qs, N),
        safe_c(Qs),
	place_queens(Qs, N).

% All predicate names below get a _c suffix
% safe_c(Qs): List Qs describes a safe placement of queens.
safe_c([]).
safe_c([Q|Qs]):- 
        no_attack_c(Qs, Q, 1),
	safe_c(Qs).

% no_attack_c(Qs, Q, I): A quuen placed in row i, position Q does not attack
% any of the queens Qs[1], Qs[2], etc, placed in rows i+I, i+I+1, etc.
no_attack_c([], _, _).
no_attack_c([X|Xs], Y, I):-
      no_threat_c(X, Y, I),
      J is I+1,
      no_attack_c(Xs, Y, J).

% Queens placed in columns X and Y in rows of distance I do not attack each
% other. A bit more formally: queens placed in row i, column X and in row
% i+I, column Y do not attack each other.
:- block no_threat_c(-, ?,?), no_threat_c(?,-,?).   % ********************
                                                   % block declaration added
no_threat_c(X, Y, I) :-
        Y =\= X, Y =\= X-I, Y =\= X+I.

compare(Ns):-
	(   number(Ns) -> N = Ns
	;   Ns = From-To -> between(From, To, N)
	;   member(N, Ns)
	),
	format('~nRunning~|~t~w~3+ queens ...', [N]), flush_output(user),
	statistics(runtime, [T0,_]),
	findall(Q, queens(N, Q), Sols),
	statistics(runtime, [T1,_]),
	length(Sols, Len),
	format('~|~t~w~5+ solutions found.~n', [Len]),
	statistics(runtime,[T2,_]),
	findall(Q, queens_c(N, Q), Sols2),
	statistics(runtime, [T3,_]),
	(   Sols == Sols2 -> true
	;   format('WARNING: solution sets are different~n~w~n~w~n~n', [Sols,Sols2])
	),
	Time1 is T1-T0,
	Time2 is T3-T2,
	format('Time:~t~3ds~15| for the generate-and-test version~n', [Time1]),
	format('Time:~t~3ds~15| for the test-with-blocking-and-generate version~n', [Time2]),
	fail.
compare(_).




%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% 2. felbontas(+X, -A, -B, -C):-
%    X, A, B, C pozitiv egesz szamok
%    X = A + B + C
%    1 < A < B < C
%    A, B, C szamok relativ primek
felbontas(X, A, B, C):-
	relPrimes(A, B), 
	relPrimes(A, C), 
	relPrimes(B, C), 
	smaller(A, B), 
	smaller(B, C), 
	osszege(A, B, C, X), 
	between(2, X, A), 
	between(2, X, B), 
	between(2, X, C).

:- block relPrimes(-, ?), relPrimes(?, -).
relPrimes(X, Y):-
	1 =:= gcd(X,Y).

:- block smaller(-, ?), smaller(?, -).
smaller(X, Y):-
	X < Y.

:- block osszege(-,-,-,?).
osszege(A, B, C, X):-
	( ground(A) -> A < X, X2 is X - A, osszege2(B, C, X2)
	; ground(B) -> B < X, X2 is X - B, osszege2(A, C, X2)
	; ground(C) -> C < X, X2 is X - C, osszege2(A, B, X2)
	).

:- block osszege2(-, -, ?).
osszege2(A, B, X):-
	( ground(A) -> B is X - A
	; ground(B) -> A is X - B
	).


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% 3. Hamming problema
%    hamming(N, H):
%    H lista az elso olyan N > 1 szamot tartalmazza novekvo sorrendben,
%    amelyek primtenyezoi kozott csak 2 es 3 szerepel.
hamming(N, H):-
        times(U, 2, X), 
        times(U, 3, Y), 
        merge(X, Y, Z),
        prefix(N, Z, H),
	U = [1|H].

% times(X, M, Z):- A Z lista az X elemeinek M-szerese
:- block times(-, ?, ?).
times([], _, []).
times([A|X], M, [B|U]):-
        freeze(A, ( B is M*A,
                    times(X, M, U)
                  )
              ).

:- block merge(-, ?, ?), merge(?, -, ?).
merge([A|X], [B|Y], [C|Z]):-
        when( (nonvar(A), nonvar(B)),
              ( A < B -> C = A, merge(X, [B|Y], Z)
              ; B < A -> C = B, merge([A|X], Y, Z)
              ; C = A, merge(X, Y, Z)
              )
            ).
merge([], X, X):- !.
merge(X, [], X).

% prefix(N, X, Y):- Az X lista elso N eleme Y.
prefix(0, _, []):- !.
prefix(N, [A|X], [A|Y]):-
        N > 0, N1 is N-1, prefix(N1, X, Y).

%    hamming(N, H, Rest):-
%    H lista az elso olyan N > 1 szamot tartalmazza novekvo sorrendben,
%    amelyek primtenyezoi kozott csak 2 es 3 szerepel.
%    Rest a "feleslegesen" kiszmolt tovbbi Hamming szmok listja
hamming(N, H, Rest):-
        times(U, 2, X), 
        times(U, 3, Y), 
        merge(X, Y, Z),
	length(H, N),
	append(H, Rest, Z),
	U = [1|H].



%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%  4. Diszjunkci
%     +(A, B, C): A es B diszjunkcioja C
+(A, B, C):-
    when( (nonvar(A); nonvar(B); ?=(A,B); nonvar(C)),
	  or(A, B, C)
         ).

% or(A, B, C): A es B diszjunkcioja C
% vagy ismerjuk az egyik argumentumot, vagy tudjuk, hogy a ket bemenet azonos    
or(A, B, C):-
    (   A == 1 -> C = 1
    ;   A == 0 -> B=C
    ;   B == 1 -> C = 1
    ;   B == 0 -> A=C
    ;	C == 0 -> A = 0, B = 0
    ;	?=(A, B) -> C = A
    ;	/* C==1 */ when( (nonvar(A); nonvar(B); ?=(A,B)), one_one(A, B))
    ).

% one_one(A, B): A es B kozul az egyik biztosan 1
one_one(A, B):-
    (   A == 1 -> true
    ;   A == 0 -> B = 1
    ;   B == 1 -> true
    ;   B == 0 -> A = 1
    ;	/* A==B */ A = 1
    ).

end_of_file.

| ?- compare(4-8).

Running  4 queens ...    2 solutions found.
Time:    0.000s for the generate-and-test version
Time:    0.000s for the test-with-blocking-and-generate version

Running  5 queens ...   10 solutions found.
Time:    0.000s for the generate-and-test version
Time:    0.000s for the test-with-blocking-and-generate version

Running  6 queens ...    4 solutions found.
Time:    0.040s for the generate-and-test version
Time:    0.000s for the test-with-blocking-and-generate version

Running  7 queens ...   40 solutions found.
Time:    0.650s for the generate-and-test version
Time:    0.010s for the test-with-blocking-and-generate version

Running  8 queens ...   92 solutions found.
Time:   14.820s for the generate-and-test version
Time:    0.050s for the test-with-blocking-and-generate version
yes
