:- use_module(library(clpq)).
:- use_module(library(lists)).
:- use_module(library(between)).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% 1.  Szlsrtk-szmols                             %
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

% dilemma(+Osszeg, -Max):
% Osszeg forinttal a zsebnkben, legfeljebb Max db boldogsgpontot tudunk
% sszegyjteni a kvetkez felttelek betartsval:
% Egy sr ra 200 Ft es 8 boldogsgpontot nyjt.
% Egy zsroskenyr ra 100 Ft es 1 boldogsgpontot nyjt.
% Viszont a sr csak akkor esik jl, ha legalbb feleannyi
% zsroskenyeret is megesznk hozza. Zsroskenyrbl trtrszt is vehetnk,
% de srt csak darabonknt rulnak.
dilemma(Osszeg, MaxPont):-
	{B = -1*(K + 8*S)},
	{S =< 2*K},
	{200*S + 100*K =< Osszeg},
	bb_inf([S],B,Inf),
	{MaxPont = -1* Inf}.


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% 2.  Egyszerstett Sudoku                            %
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

% sudoku_simple(?Matrix, +N):
% Matrix egy N*N-es mtrix amely pozitv, N-nl nem nagyobb szmokbl ll.
% Minden sorban s oszlopban a szmok pronknt klnbzek.
% Minden sor s oszlop sszege N*(N+1)/2.
% Az eljrs ne hozzon ltre vlasztsi pontot!
sudoku_simple(Matrix, N):-
        length(Matrix, N),
	{Sum = N*(N+1)/2},
	korlatoz_sorok(Matrix, N, Sum),
        transpose(Matrix, TMatrix),
	korlatoz_sorok(TMatrix, N, Sum).

% korlatoz_sorok(+Sorok, +N, +Sum):
% Sorok egy K*N-es mtrix (K>=0), amely pozitv,
% N-nl nem nagyobb szmokbl ll.
% Minden sorban a szmok pronknt klnbzek.
% Minden sor sszege Sum.
korlatoz_sorok([],_,_).
korlatoz_sorok([Sor|Sorok], N, Sum):-
	length(Sor, N),
	osszege(Sor, Sum),
	kulonbozoek_0_N(Sor, N),
	korlatoz_sorok(Sorok, N, Sum).

% osszege(L, Sum): Az L lista sszege Sum.
osszege([], Sum):-
	{Sum = 0}.
osszege([X|Xs], Sum):-
	{Sum = Sum1 + X},
	osszege(Xs, Sum1).

% kulonbozoek_0_N(L, N): Az L lista elemei pozitv, N-nl nem nagyobb,
% egymstl klnbz szmok.
kulonbozoek_0_N([],_).
kulonbozoek_0_N([X|Xs], N):-
	{X > 0, X =< N},
	kulonboznek_tole(Xs, X),
	kulonbozoek_0_N(Xs, N).

% kulonboznek_tole(L, X): Az L lista elemei klnbznek X-tl.
kulonboznek_tole([], _).
kulonboznek_tole([X|Xs], Y):-
	{ X =\= Y },
	kulonboznek_tole(Xs, Y).


% Mx minden eleme 1 s N kz es egsz szm (CLPQ brzolsban!).
place_sudoku(Mx, N) :-
	(   foreach(R, Mx), param(N)
	do  (   foreach(E, R), param(N)
	    do  between(1, N, V), {E=V}
	    )
	).

% place_sudoku1(Mx, N): ugyanaz mint place_sudoku(Mx, N).
place_sudoku1([], _).
place_sudoku1([R|Mx], N) :-
	place_row(R, N),
	place_sudoku1(Mx, N).

% place_row(L, N):
% Az L lista  minden eleme 1 s N kz es egsz szm (CLPQ brzolsban!).
place_row([], _).
place_row([E|R], N) :-
	between(1, N, V),
	{E = V},
	place_row(R, N).


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% 3.  Fibonacci                                        %
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

% fibonacci0(?L, +N, +Max):
% L pozitv szmok N hossz listja, N >= 3.
% Az L lista elemei nem lehetnek nagyobbak Max-nl.
% A lista brmely hrom A,B,C egymst kvet elemre teljesl,
% hogy C = A + B.
% Az eljrs ne hozzon ltre vlasztsi pontot!
fibonacci0(L, N, Max):-
        length(L,N),
        correct_sequence(L, Max).

% correct_sequence(L, Max):
% Az L lista elemei pozitvak s nem lehetnek nagyobbak Max-nl.
% A lista brmely hrom A,B,C egymst kvet elemre teljesl,
% hogy C = A + B.
correct_sequence([A|Tail], Max):- 
	{A > 0, A =< Max}, Tail = [B|Tail2],
	(   Tail2 == [] -> {B > 0, B =< Max}
	;   Tail2 = [C|_], {C = A + B},
	    correct_sequence(Tail, Max)
	).

% fibonacci(?L, +N, +Max):
% L pozitv EGSZ szmok N hossz listja, N >= 3.
% Az L lista elemei nem lehetnek nagyobbak Max-nl.
% A lista brmely hrom A,B,C egymst kvet elemre teljesl,
% hogy C = A + B.
% Az eljrs felsorolja az sszes megoldst.
fibonacci(L, N, Max):-
	fibonacci0(L, N, Max),
	L = [A,B|_],
	bb_inf(L, -A, MinMA), 
	bb_inf(L, -B, MinMB),
	between_clp(1, -MinMA, A),
	between_clp(1, -MinMB, B).

% N =< X =< M, X egsz.
between_clp(N, M, X) :- {N =< M},
	(   {X = N}
	;   {N  < M},
	    between_clp(N+1, M, X)
	).

% Var egsz szm, feltve, hogy L elemei mind egszek.
% Visszalps sorn Var felvesz minden lehetsges rtket.
clpq_label(L, Var) :-
	bb_inf([Var|L], Var, Min),
	bb_inf([Var|L], -Var, NegMax),
	Min = rat(MinI,1),
	NegMax = rat(NegMaxI,1),
	MaxI is -NegMaxI,
	between(MinI, MaxI, Int),
	{Var = Int}.

% Azonos fibonacci/3-mal.
fibonacci1(L, N, Max):-
	fibonacci0(L, N, Max),
	L = [A,B|_],
	clpq_label(L, A),
	clpq_label(L, B).

% fibonacci(N, F): F a klasszikus Fibonacci sorozat (1 1 2 3 5 ...)
% N. tagja, 1-tl szmozva, N >= 3.
fibonacci(N, F) :-
	fibonacci0(L, N, F), bb_inf(L, F, F).


