:- use_module(library(clpfd)).
:- use_module(library(lists)).


% 0. Tekintsd az albbi kzismert SEND MORE MONEY kriptaritmetikai feladvnyt:

%    Az albbi egyenletben egyes betket szmjegyekkel kell helyettesteni,
%    gy hogy az albbi sszeads helyes legyen. Egy bet minden
%    elfordulst ugyanazzal a szmjeggyel kell helyettesteni, klnbz
%    betket klnbz szmjegyekkel, szmeleji 0 nem megengedett.

% 			    SEND + MORE = MONEY
  
% a. Fogalmazd t ezt a feladvnyt egy <X, D, C> CSP feladatt. A kvetkez
%    relcijeleket (korltokat) hasznld:

%    p(X1, X2, thozat, tvitel, X3): X1+X2+thozat = tvitel*10+X3
%    p(X1, X2, tvitel, X3):          X1+X2 = tvitel*10+X3
%    d(X1, X2):	      		      X1 \= X2	 

%    Itt Xi 0-9, thozat, tvitel 0-1 rtkeket vehet fel,

% b. A felesleges vltozrtkek elhagysval kszitsd el az elbbi CSP
%    feladat lkonzisztens vltozatt!

% c, Vlassz egy olyan vltozt, amelynek cmkzsvel az sszes vltoz
%    rtket kap s gy elll az els (s egyetlen) megolds!

/*   Megolds:

a.   
     X           D
     ---------------------
     S           {1,...,9}
     E           {0,...,9}
     N           {0,...,9}
     D           {0,...,9}
     M           {1,...,9}
     O           {0,...,9}
     R           {0,...,9}
     Y           {0,...,9}
     C1          {0,1}
     C2          {0,1}
     C3          {0,1}

     C:
     d(S,E), d(S,N), ..., d(S,Y), d(E,N), d(E,D), ..., d(E,Y), ..., d(R,Y),
     % sszesen (8 alatta 2) d/2 korlt van
     %           M
     p(S,M,C3,M, O),
     p(E,O,C2,C3,N),
     p(N,R,C1,C2,E), 
     p(D,E,   C1,Y).

     ahol p/4 = {<0,0,0,0>,
                 <0,1,0,1>,
		 ...
                 <0,9,0,9>,
                 <1,0,0,1>,
                 <1,1,0,2>,
                 <1,2,0,3>,
                 <1,9,1,0>,
		 ...
                 <9,9,1,8>}
     % sszesen 100 szmngyes tartozik p/4-hez.

     p/5 analg mdon adhat meg, 200 szmtst tartalmaz.

b. A fenti CSP feladatbl a felesleges vltozrtkek elhagysval elll
   lkonzisztens CSP az albbi:

     X           D
     ---------------------
     S           {9}
     E           {2,...,8}
     N           {2,...,8}
     D           {2,...,8}
     M           {1}
     O           {0}
     R           {2,...,8}
     Y           {2,...,8}
     C1          {0,1}
     C2          {0,1}
     C3          {0}

     C: A biztosan teljesl korltok elhagysa utn ez marad:

     d(E,N),
     d(E,D),
     d(E,R),
     d(E,Y),
     d(N,D),
     d(N,R),
     d(N,Y),
     d(D,R),
     d(D,Y),
     d(R,Y),
     p(E,O,C2,C3,N),
     p(N,R,C1,C2,E), 
     p(D,E,   C1,Y)

c.
     Az E vagy N  vltoz cmkzsvel az sszes vltoz
     rtket kap s elll az els (s egyetlen) megolds:

     X           D
     ---------------------
     S           {9}
     E           {5}
     N           {6}
     D           {7}
     M           {1}
     O           {0}
     R           {8}
     Y           {2}
     C1          {1}
     C2          {1}
     C3          {0}

     A D, R vagy Y vltoz (illetve a C1, C2 vltozk valamelyiknek)
     cmkzsekor nem helyettestdik be az sszes vltoz.
 

*/


% 1. % smm(-L): L = [S,E,N,D,M,O,R,Y] a SEND MORE MONEY feladvny megoldsa.

%    | ?- smm(L).

%    L = [9,5,6,7,1,0,8,2] ? ; no

% Az L = [S,E,N,D,M,O,R,Y] s Cs = [C1,C2,C3] vltozkra fennll a
% SEND+MORE=MONEY korlt-feladat. Az eljrs nem cimkz.
smm_base(L, Cs) :-
	L = [S,E,N,D,M,O,R,Y],
	Cs = [C1,C2,C3],
	domain(L, 0, 9), S #\= 0, M #\= 0,
	all_diff(L),
	%           M
	p(S,M,C3,M, O),
	p(E,O,C2,C3,N),
	p(N,R,C1,C2,E), 
	p(D,E,   C1,Y).

smm(L) :-
	smm_base(L, _),
	labeling([], L).

% all_diff(L): az L vltozlista elemei pronknt klnbzek.
all_diff([]).
all_diff([X|L]) :-
	out_of(L, X),
	all_diff(L).

% out_of(L, Y): az L vltozlista minden eleme klnbzik Y-tl.
out_of([], _).
out_of([X|L], Y) :-
	X #\= Y,
	out_of(L, Y).

% X1+X2+thozat #= tvitel*10+X3.
p(X1, X2, thozat, tvitel, X3) :-
	scalar_product([1,1,1,-10], [X1,X2,thozat,tvitel], #=, X3,
		       [consistency(domain)]).

% X1+X2 #= tvitel*10+X3.
p(X1, X2, tvitel, X3) :-
	scalar_product([1,1,-10], [X1,X2,tvitel], #=, X3,
		       [consistency(domain)]).

	


% 2. % sudoku_simple(?Matrix, +N):
%    % Matrix egy N*N-es mtrix amely 1 s N kz es szmokbl ll.  Minden
%    % sorban s oszlopban a szmok pronknt klnbzek.   

%    % Az eljrs ne hozzon ltre vlasztsi pontot (ne cmkzzen)!

%    | ?- Mx = [[3,1|_]|_], sudoku_simple(Mx, 3), append(Mx, _V), labeling([],_V).
%    Mx = [[3,1,2],[1,2,3],[2,3,1]] ? ;
%    Mx = [[3,1,2],[2,3,1],[1,2,3]] ? ; no

% 2. % sudoku_simple(?Matrix, +N):
%    % Matrix egy N*N-es mtrix amely 1 s N kz es szmokbl ll.  Minden
%    % sorban s oszlopban a szmok pronknt klnbzek.   

%    % Az eljrs ne hozzon ltre vlasztsi pontot (ne cmkzzen)!

%    | ?- Mx = [[3,1|_]|_], sudoku_simple(Mx, 3), append(Mx, _V), labeling([],_V).
%    Mx = [[3,1,2],[1,2,3],[2,3,1]] ? ;
%    Mx = [[3,1,2],[2,3,1],[1,2,3]] ? ; no


sudoku_simple(Matrix, N) :-
        length(Matrix, N),
        all_lengths(Matrix, N),
        dom_constraint(Matrix, N),
        transpose(Matrix, ColMatrix),
        dom_constraint(ColMatrix, N),
	append(Matrix, List),
	labeling([], List).

% all_lengths(Mx, Len): Az Mx lista minden eleme egy Len hossz lista.
all_lengths([], _).
all_lengths([L|Ls], Len) :-
	length(L, Len),
	all_lengths(Ls, Len).


% dom_constraint(Mx, N): Az Mx lista minden L elemre fennll, hogy L az
% 1..N egszek egy permutcijt tartalmazza.
dom_constraint([], _).
dom_constraint([L|Ls], N) :-
        domain(L, 1, N),
        all_distinct(L),
        dom_constraint(Ls, N).


% 3. Bvs ngyzetnek neveznk egy N oldalhossz ngyzetes mtrixot, ha az az
%    1..N*N szmokat pontosan egyszer tartalmazza, minden sor, oszlop s tl
%    sszege ugyanaz.

%    % magic(+N, ?Mx): Mx egy N oldalhossz bvs ngyzet

%    | ?- magic(3, LL).
%    LL = [[2,7,6],[9,5,1],[4,3,8]] ? ;
%    LL = [[2,9,4],[7,5,3],[6,1,8]] ? ;
%    LL = [[4,3,8],[9,5,1],[2,7,6]] ? ;
%    LL = [[4,9,2],[3,5,7],[8,1,6]] ? ;
%    LL = [[6,1,8],[7,5,3],[2,9,4]] ? ;
%    LL = [[6,7,2],[1,5,9],[8,3,4]] ? ;
%    LL = [[8,1,6],[3,5,7],[4,9,2]] ? ;
%    LL = [[8,3,4],[1,5,9],[6,7,2]] ? ;
%    no

magic(N, Mx) :-
	Max is N*N,
	length(Mx, N), 
	all_lengths(Mx, N),
	append(Mx, List),
	domain(List, 1, Max),
	all_distinct(List),
	transpose(Mx, TMx),
	diagonal(Mx, 0, Diag1),
	all_reverse(Mx, RMx),
	diagonal(RMx, 0, Diag2),
	append([Mx, TMx, [Diag1,Diag2]], AllToSum),
	(   foreach(L, AllToSum), param(Sum)
	do  sum(L, #=, Sum)
	),
	labeling([], List).

% all_reverse(Mx, RMx): Az Mx mtrix minden sornak megfordtsval ll el
% az RMx mtrix.
all_reverse(Mx, RMx) :-
	(   foreach(L, Mx), foreach(RL, RMx)
	do  reverse(L, RL)
	).

% diagonal(Mx, N, D): D is the diagonal of the matrix obtained from Mx
% by omitting its first N columns.
diagonal([Row|Mx], N, [D|Ds]) :-
	nth0(N, Row, D),
	N1 is N+1,
	diagonal(Mx, N1, Ds).
diagonal([], _, []).


% 4. rd meg az albbi p(A,B,C) korltot!

%    % p(A,B,C): ha az A-B szm 3-mal osztva 1 maradkot ad, akkor C pros,
%    % klnben pratlan.
%    % A p/3 eljrs ne cmkzzen! 

%    Prbld meg megrni ezt a korltot reifikci nlkl is!

%    | ?- domain([A,B], 1, 3), C in 0..1, p(A, B, C), 
%         labeling([], [A,B,C]), write(A-B-C), nl, fail.
%    1-1-1
%    1-2-1
%    1-3-0
%    2-1-0
%    2-2-1
%    2-3-1
%    3-1-1
%    3-2-0
%    3-3-1
%    no

p(A, B, C) :-
	(A-B) mod 3 #= 1 #<=> C mod 2 #= 0.

% Reifikci nlkl:

p2(A, B, C) :-
	((A-B) mod 3 + 1 - C) mod 2 #= 0.


% 5. Egy szmlistban loklis szlsrtknek hvunk egy elemet, ha mindkt
%    szomszdjnl hatrozottan nagyobb, vagy mindkettnl hatrozottan kisebb.

%    % szeszam(+L, ?K): az L, csupa klnbz elembl ll listban
%    % lev loklis szlsrtkek szma K.  
%    % Az eljrs ne hozzon ltre vlasztsi pontot (ne cmkzzen)!

%    | ?- L=[1,_,_,_], domain(L, 1, 4), szeszam(L, 2), labeling([], L).
%    L = [1,3,2,4] ? ;
%    L = [1,4,2,3] ? ; no

szeszam(L, K) :- 
	all_distinct(L),
	(   L = [A,B|L1] -> 
	    szeszam(L1, A, B, K)
	;   K = 0
	).

% szeszam(L, A, B, K): Az [A,B|L] listban K darab loklis szlsrtk van.
szeszam([C|L], A, B, K) :-
	(A #< B #/\ C #< B) #\/ (A #> B #/\ C #> B) #<=> K1,
	K #= K1 + K0,
	szeszam(L, B, C, K0).
szeszam([], _, _, 0).


% 6. Egy szmlistban balrl lthatnak hvunk egy elemet, ha az
%    hatrozottan nagyobb az sszes t megelz elemnl.

%    % latszam(+L, ?K): az L listban lev balrl lthat elemek szma K.  
%    % Az eljrs ne hozzon ltre vlasztsi pontot (ne cmkzzen)!

%    | ?- L=[_,_,2,_], domain(L, 1, 4), all_distinct(L), latszam(L, 3), 
%         labeling([], L).
%    L = [1,3,2,4] ? ; no

latszam(L, K):-
        latszam(L, 0, K).

% latszam(L, Max, K): az L listban lev, Max-nl nagyobb lthat elemek
% szma K.
latszam([], _, 0).
latszam([L|Ls], Max, K):-
        L #> Max #<=> B,
        Max2 #= max(L, Max),
	K #= K0+B,
        latszam(Ls, Max2, K0).


% 7. 

%    % panorama(+N, +Latvanyok, ?Lakotelep):
%    % Lakotelep egy N*N-es mtrix, amely egy laktelep alaprajzt adja
%    % ki. A mtrix elemei az egyes pletek magassgt mutatjk. Minden
%    % sorban es minden oszlopban klnbz magassgak az pletek s ezek
%    % a magassgok az 1..N tartomnybl kerlnek ki. Latvanyok egy olyan
%    % lista, amelynek elemei bal(I,K), felul(J,K), jobb(I,K), alul(J,K)
%    % alak Prolog kifejezsek, ahol I, J s K egyarnt az 1..N
%    % intervallumba esik. A Latvanyok listban elfordul bal(I,K) elem
%    % azt a korltozst fejezi ki, hogy a laktelep I-edik sort balrl
%    % nzve K hz ltszik, a felul(J,K) azt, hogy a laktelep J-edik
%    % oszlopt fellrl nzve K hz ltszik, st.

%    % Az eljrs sorolja fel az sszes megoldst, gy hogy csak a
%    % labeling/2 knyvtri eljrs hvsa hoz ltre vlasztsi pontot!

%    | ?- panorama(4, [bal(1,2),bal(3,2),felul(2,3),felul(4,3),
%                      jobb(2,3),jobb(4,1),alul(1,3),alul(3,3)], Mx).
%    Mx = [[3,1,4,2],[4,2,3,1],[2,4,1,3],[1,3,2,4]] ? ; no

panorama(N, Latvanyok, Lakotelep):-
	sudoku_simple(Lakotelep, N),
	transpose(Lakotelep, TLakotelep),
	all_reverse(Lakotelep, RLakotelep),
	all_reverse(TLakotelep, RTLakotelep),
	Param = [bal-Lakotelep,jobb-RLakotelep,
		 felul-TLakotelep,alul-RTLakotelep],
	latvanyok(Latvanyok, Param),
        append(Lakotelep, AllVars),
        labeling([], AllVars).

latvanyok([], _Param).
latvanyok([Lv|Lvk], Param) :-
	Lv =.. [Irany,I,K],
	memberchk(Irany-Mx, Param),
	nth1(I, Mx, L),
	latszam(L, K),
	latvanyok(Lvk, Param).

% 8. % square(+N, +XSumList, +YSumList, ?Res):- Res egy N*N-es mtrix, N>=3,
%    % XSumList valamint YSumList pedig N hossz listk. A Res mtrix elemei
%    % az 1..N intervallumba esnek, s teljesl rjuk, hogy a (vzszintesen
%    % vagy fgglegesen) oldalszomszdos elemek mindig klnbzek. Res
%    % sorainak sszegt rendre XSumList elemei adjk. Hasonlan, Res
%    % oszlopainak sszegei YSumList-ben vannak felsorolva. Tovbb, ha
%    % brmely sorbl illetve oszlopbl kivlasztunk hrom egymst kvet
%    % (A,B,C) elemet, akkor ezekre teljesl az albbi tulajdonsg: ha az A-B 
%    % szm 3-mal osztva 1 maradkot ad, akkor C pros, klnben pratlan.

%    % Az eljrs sorolja fel az sszes megoldst, gy hogy csak a
%    % labeling/2 knyvtri eljrs hvsa hoz ltre vlasztsi pontot!

%    | ?- square(4, Res, [12,_,_,8], [6,_,_,_]).
%    Res = [[1,4,3,4],[2,3,1,3],[1,4,3,4],[2,1,2,3]] ? ;
%    Res = [[1,4,3,4],[2,3,1,3],[1,4,3,4],[2,1,4,1]] ? ; no

square(N, Res, XSumList, YSumList):-
	length(Res, N),
	all_lengths(Res, N),
	append(Res, All),
	domain(All, 1, N),
	transpose(Res, TransRes),
	sum_constraints(Res, XSumList),
	sum_constraints(TransRes, YSumList),
	append(Res, TransRes, Lines),
	constrain_list(Lines),
	labeling([], All).

sum_constraints([], []).
sum_constraints([L|Ls], [Sum|Sums]):-
	sum(L, #=, Sum),
	sum_constraints(Ls, Sums).

constrain_list([]).
constrain_list([L|Ls]):-
	L = [A,B|L1],
	constrain(L1, A, B),
	constrain_list(Ls).

constrain(L, A, B) :-
	A #\= B,
	(   L == [] -> true
	;   L = [C|L1],
	    p(A, B, C),
	    constrain(L1, B, C)
	).

