:- use_module(library(clpfd)).
:- use_module(library(lists)).

:- multifile clpfd:dispatch_global/4.
:- discontiguous clpfd:dispatch_global/4.

/*

Szktsi lehetsgek egy tobb1mint2(L) hvs esetn
----------------------------------------------------

Az Xs korltvltoz-lista minden X elemre definiljuk a kvetkez
tulajdonsgokat (k = 1, 2):

    X biztosan k <=> ha dom(X) = {k} (azaz X == k)
    X lehet k    <=> ha k eleme dom(X), de dom(X) \= {k}

Egy tobb1mint2(Xs) globlis korlt a kvetkezkppen kell viselkedjen:

    Vezessk be a kvetkez jellseket (k = 1, 2):

	Bk = az Xs listban lev "biztosan k" tulajdonsg elemek szma
	Lk = az Xs listban lev "lehet k" tulajdonsg elemek szma

    Ha B1 > B2+L2                                                        (1)
    => sikeres lefuts, szkts nlkl

    Ha B2 >= B1+L1                                                       (2)
    => sikertelen lefuts

    Ha B2+1 = B1+L1                                                      (3)
    => sikeres lefuts, gy, hogy

      (a) minden olyan X, amely lehet 1, az 1 rtkre szkl
    
      (b) minden olyan X tartomnybl, amely lehet 2, elhagyjuk a 2 rtket
    
    (Ha egy (a)-beli X nem 1 rtk lenne, akkor a 1-esek maximlis
    szma B1+L1-1 = B2 lenne, azaz a korlt nem llhatna fenn;
    hasonlkppen ha egy (b)-beli X 2 rtk lenne, akkor a biztos 2-esek
    szma B2+1 = B1+L1-re nne, azaz a korlt ilyenkor sem llhatna fenn;
    viszont az (a) s (b) szktsekkel a korlt mr biztosan fennll,
    teht nincs szksg tovbbi szktsre, ezrt a globlis korlt
    sikerrel kilphet)

    Ha B2+2 = B1+L1                                                       (4)
    => minden olyan X tartomnybl, amely lehet 1 s lehet 2 is,
    elhagyjuk a 2 rtket    
    (Ha egy ilyen vltoz 2-es rtket kapna, akkor a biztos 2-esek szma
    B2+1 = B1+L1-1-re nne, a 1-esek maximlis szma viszont ugyanennyire,
    B1+L1-1-re cskkenne, azaz a korlt mr biztosan nem llhatna fenn.
    Ez a szkts mg nem garantlja a korlt teljeslst, teht nem
    lphetnk ki a globlis korltbl)

    Minden egyb esetben szkts nlkl jra elaltatjuk a globlis korltot

A globlis korlt llapot-adatstruktrja
-----------------------------------------

Az llapot egy Xs-ET pr, ahol

    Xs a korlt lista-argumentuma, amelybl esetleg elhagytuk a biztos 1-es s
    biztos 2-es elemeket, valamint azokat, amelyek tartomnya nem tartalmazza
    sem az 1-es sem a 2-es rtket.
    
    ET (1-esek tbblete) egy olyan (esetleg negatv) egsz szm, amely az
    elhagyott biztos 1-esek s az elhagyott biztos 2-esek szmnak klnbsgt
    tartalmazza.

Az llapot szrse
------------------

% filter(+Xs0, -Xs, +ET0, -ET, -X1s, -X2s, -X12s):

A filter segdeljrs 1. s 3. argumentuma egy llapot kt rsze. Az ltala
visszaadott Xs s ET rtkekre mr garantlt, hogy Xs nem tartalmaz biztos
1-es, sem biztos 2-es elemeket (s persze ET rtke figyelembe veszi az gy
elhagyott elemeket), tovbb Xs olyan elemeket sem tartalmaz, amelyek
tartomnyban 1-es s 2-es rtk sem szerepel.

A filter emellett a visszaadott Xs listt hrom diszjunkt rszre
bontva is ellltja:

X1s:  Xs azon elemei amelyek lehetnek 1-esek, de nem lehetnek 2-esek
X2s:  Xs azon elemei amelyek lehetnek 2-esek, de nem lehetnek 1-esek
X12s: Xs azon elemei amelyek lehetnek 1-esek s lehetnek 2-esek is

Jelljk ezen listk hosszt rendre N1, N2, N12-vel. A korbban bevezetett
B1, B2, L1, L2 rtkek az albbi mdon kapcsoldnak az ET, N1, N2, N12
rtkekhez:

    ET = B1-B2
    L1 = N1+N12
    L2 = N2+N12

Ennek megfelelen a korbbi (1)-(4) felttelek az albbi alakot ltik:



       B1 > B2+L2                                                        (1)
                               ET > N2+N12                               (1a)

       B2 >= B1+L1                                                       (2)
                               -ET >= N1+N12                             (2a)

       B2+1 = B1+L1                                                      (3)
                               -ET = N1+N12-1                            (3a)

       B2+2 = B1+L1                                                      (4)
                               -ET = N1+N12-2                            (4a)

*/

tobb1mint2(Xs) :-
	(   foreach(X, Xs), foreach(dom(X), Susps)
	do  true
	),
        fd_global(tobb1mint2(Xs), Xs-0, Susps).


clpfd:dispatch_global(tobb1mint2(_), Xs0-ET0, Xs-ET, Act) :- 
        filter(Xs0,Xs, ET0,ET, X1s, X2s, X12s),
	length(X1s, N1),
	length(X2s, N2),
	length(X12s, N12),
        (   ET > N2+N12            % biztosan tbb, lsd (1)
        ->  Act = [exit]
        ;   -ET >= N1+N12          % biztosan nem tbb, lsd (2)
        ->  Act = [fail]
	;   -ET >= N1+N12-1        % ha minden lehetsges 1-et 1-re
	                           % helyettestnk, s egy lehetsges 
	                           % 2-es sem lesz 2-es, akkor pp,
				   % hogy csak 1-gyel tbb lesz, lsd (3)
	->  all_in_range(X1s, {1}, Act, Act1),
	    all_in_range(X12s, {1}, Act1, Act2),
	    all_in_range(X2s, \{2}, Act2, [exit])
	;   -ET >= N1+N12-2        % 2-es nem "vehet el helyet" 1-estl (4)
	->  all_in_range(X12s, \{2}, Act, [])
        ;   Act = []
        ).

filter([],[], ET,ET, [], [], []).
filter([X|Xs], Ys, ET0,ET, Y1s, Y2s, Y12s) :-
	(   X == 1 -> ET1 is ET0+1, filter(Xs, Ys, ET1,ET, Y1s, Y2s, Y12s)
	;   X == 2 -> ET1 is ET0-1, filter(Xs, Ys, ET1,ET, Y1s, Y2s, Y12s)
	;   fd_set(X, FDS),
	    (   fdset_member(1, FDS)
	    ->  (   fdset_member(2, FDS)
		->  Y12s = [X|NY12s],
		    Ys = [X|NYs],
		    filter(Xs, NYs, ET0,ET, Y1s, Y2s, NY12s)
		;   Y1s = [X|NY1s],
		    Ys = [X|NYs],
		    filter(Xs, NYs, ET0,ET, NY1s, Y2s, Y12s)
		)
	    ;   fdset_member(2, FDS)
	    ->  Y2s = [X|NY2s],
		Ys = [X|NYs],
		filter(Xs, NYs, ET0,ET, Y1s, NY2s, Y12s)
	    ;   filter(Xs, Ys, ET0,ET, Y1s, Y2s, Y12s)
	    )
	).

all_in_range(L, FDRange, Act0, Act) :-
	range_to_fdset(FDRange, FDS),
	all_in_fdset(L, FDS, Act0, Act).


all_in_fdset([], _, A0, A0).
all_in_fdset([X|L], FDS, [X in_set FDS|A1], A) :-
	all_in_fdset(L, FDS, A1, A).
